"""
Backfill the new remote_created_at and remote_updated_at content audit record values.
"""
import logging

from django.apps import apps
from django.contrib import auth
from django.core.management.base import BaseCommand
from django.db.models import Q

from integrated_channels.integrated_channel.management.commands import IntegratedChannelCommandMixin
from integrated_channels.utils import batch_by_pk, generate_formatted_log

User = auth.get_user_model()

LOGGER = logging.getLogger(__name__)


class Command(IntegratedChannelCommandMixin, BaseCommand):
    """
    Update content transmission items to have the new remote_created_at and remote_updated_at values.

    ./manage.py lms backfill_remote_action_timestamps
    """

    def handle(self, *args, **options):
        """
        Update all past content transmission items remote_created_at and remote_updated_at
        """

        ContentMetadataItemTransmission = apps.get_model(
            'integrated_channel',
            'ContentMetadataItemTransmission'
        )

        no_remote_created_at = Q(remote_created_at__isnull=True)
        for items_batch in batch_by_pk(ContentMetadataItemTransmission, extra_filter=no_remote_created_at):
            for item in items_batch:
                try:
                    item.remote_created_at = item.created
                    item.remote_updated_at = item.modified
                    item.save()
                    LOGGER.info(generate_formatted_log(
                        item.integrated_channel_code,
                        item.enterprise_customer.uuid,
                        None,
                        item.content_id,
                        f'ContentMetadataItemTransmission <{item.id}> '
                        f'remote_created_at={item.remote_created_at}, '
                        f'remote_updated_at={item.remote_updated_at}'
                    ))
                except Exception:  # pylint: disable=broad-except
                    LOGGER.exception(generate_formatted_log(
                        item.integrated_channel_code,
                        item.enterprise_customer.uuid,
                        None,
                        item.content_id,
                        f'ContentMetadataItemTransmission <{item.id}> '
                        'error backfilling remote_created_at & remote_updated_at'
                    ))



# Generated by Django 2.2.19 on 2021-02-22 22:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('canvas', '0006_canvaslearnerassessmentdatatransmissionaudit'),
    ]

    operations = [
        migrations.AlterField(
            model_name='canvasenterprisecustomerconfiguration',
            name='canvas_account_id',
            field=models.BigIntegerField(help_text='Account number to use during api calls. Called account_id in canvas.  Required to create courses etc.', null=True, verbose_name='Canvas Account Number'),
        ),
        migrations.AlterField(
            model_name='historicalcanvasenterprisecustomerconfiguration',
            name='canvas_account_id',
            field=models.BigIntegerField(help_text='Account number to use during api calls. Called account_id in canvas.  Required to create courses etc.', null=True, verbose_name='Canvas Account Number'),
        ),
    ]



"""
Serializers for Canvas.
"""
from integrated_channels.api.serializers import EnterpriseCustomerPluginConfigSerializer
from integrated_channels.canvas.models import CanvasEnterpriseCustomerConfiguration


class CanvasEnterpriseCustomerConfigurationSerializer(EnterpriseCustomerPluginConfigSerializer):
    class Meta:
        model = CanvasEnterpriseCustomerConfiguration
        extra_fields = (
            'client_id',
            'client_secret',
            'canvas_account_id',
            'canvas_base_url',
            'refresh_token',
            'uuid',
            'oauth_authorization_url',
        )
        fields = EnterpriseCustomerPluginConfigSerializer.Meta.fields + extra_fields



"""
Views for the ``enterprise-customer-sso-configuration`` API endpoint.
"""

import re
from xml.etree.ElementTree import fromstring

import requests
from edx_rbac.decorators import permission_required
from rest_framework import permissions, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.status import (
    HTTP_200_OK,
    HTTP_201_CREATED,
    HTTP_400_BAD_REQUEST,
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND,
)

from django.contrib import auth
from django.core.exceptions import FieldDoesNotExist, ValidationError
from django.db import transaction

from enterprise import models
from enterprise.api.utils import get_enterprise_customer_from_user_id
from enterprise.api.v1 import serializers
from enterprise.api_client.sso_orchestrator import SsoOrchestratorClientError
from enterprise.constants import ENTITY_ID_REGEX
from enterprise.logging import getEnterpriseLogger
from enterprise.models import EnterpriseCustomer, EnterpriseCustomerSsoConfiguration, EnterpriseCustomerUser
from enterprise.tasks import send_sso_configured_email
from enterprise.utils import localized_utcnow

User = auth.get_user_model()

LOGGER = getEnterpriseLogger(__name__)

BAD_CUSTOMER_ERROR = 'Must provide valid enterprise customer'
CONFIG_UPDATE_ERROR = 'Error updating SSO configuration record'
CONFIG_CREATE_ERROR = 'Error creating SSO configuration record'
BAD_IDP_METADATA_URL = 'Must provide valid IDP metadata url'


class EnterpriseCustomerInactiveException(Exception):
    """
    Exception raised when an enterprise customer is inactive.
    """


class SsoConfigurationApiError(requests.exceptions.RequestException):
    """
    Exception raised when the Sso configuration api encounters an error while fetching provider metadata.
    """


class EntityIdNotFoundError(Exception):
    """
    Exception raised by the SSO configuration api when it fails to fetch a customer IDP's entity ID from the metadata.
    """


def check_user_part_of_customer(user, enterprise_customer):
    """
    Checks if a user is in an enterprise customer.
    """
    ent_customer = enterprise_customer.enterprise_customer_users.filter(user_id=user.id)
    if ent_customer and not ent_customer.first().active:
        raise EnterpriseCustomerInactiveException
    return ent_customer.exists() or user.is_staff


def get_customer_from_request(request):
    """
    Gets the enterprise customer from the request or the user.
    """
    if customer_uuid := request.query_params.get('enterprise_customer'):
        return customer_uuid
    return get_enterprise_customer_from_user_id(request.user.id)


def fetch_configuration_record(kwargs):
    """
    Fetches the configuration record for the given uuid.
    """
    return EnterpriseCustomerSsoConfiguration.all_objects.filter(pk=kwargs.get('configuration_uuid'))


def get_metadata_xml_from_url(url):
    """
    Gets the metadata xml from the given url.
    """
    response = requests.get(url)
    if response.status_code >= 300:
        raise SsoConfigurationApiError(f'Error fetching metadata xml from provided url: {url}')
    return response.text


def fetch_entity_id_from_metadata_xml(metadata_xml):
    """
    Fetches the entity id from the metadata xml.
    """
    root = fromstring(metadata_xml)
    if entity_id := root.get('entityID'):
        return entity_id
    elif entity_descriptor_child := root.find('EntityDescriptor'):
        return entity_descriptor_child.get('entityID')
    else:
        # find <EntityDescriptor entityId=''> and parse URL from it
        match = re.search(ENTITY_ID_REGEX, metadata_xml, re.DOTALL)
        if match:
            return match.group(2)
    raise EntityIdNotFoundError('Could not find entity ID in metadata xml')


def fetch_request_data_from_request(request):
    """
    Helper method to fetch the request data dictionary from the request object.
    """
    if hasattr(request.data, 'dict'):
        return request.data.dict().copy()
    return request.data.copy()


class EnterpriseCustomerSsoConfigurationViewSet(viewsets.ModelViewSet):
    """
    API views for the ``EnterpriseCustomerSsoConfiguration`` model.
    """
    permission_classes = (permissions.IsAuthenticated,)
    queryset = models.EnterpriseCustomerSsoConfiguration.objects.all()

    serializer_class = serializers.EnterpriseCustomerSsoConfiguration

    # ``can_manage_enterprise_orchestration_configs`` maps to the edx operator system wide role. Meaning only operators
    # can complete orchestration record configuration process. The only intended caller for this endpoint is the SSO
    # orchestration api.
    @permission_required(
        'enterprise.can_manage_enterprise_orchestration_configs',
    )
    @action(methods=['post'], detail=True)
    def oauth_orchestration_complete(self, request, configuration_uuid, *args, **kwargs):
        """
        SSO orchestration completion callback. This endpoint is called by the SSO orchestrator when it has completed
        the configuration process.
        """
        # Make sure the config record exists
        sso_configuration_record = self.queryset.filter(pk=configuration_uuid).first()
        if not sso_configuration_record:
            return Response(status=HTTP_404_NOT_FOUND)

        # Small validation logging to ensure data integrity
        if not sso_configuration_record.submitted_at:
            LOGGER.warning(
                f'SSO configuration record {sso_configuration_record.pk} has received a completion callback but has'
                ' not been marked as submitted.'
            )

        if error_msg := request.POST.get('error'):
            LOGGER.error(
                f'SSO configuration record {sso_configuration_record.pk} has failed to configure due to {error_msg}.'
            )
            sso_configuration_record.errored_at = localized_utcnow()
            sso_configuration_record.save()
            return Response(status=HTTP_200_OK)

        # Mark the configuration record as active IFF this the record has never been configured.
        if not sso_configuration_record.configured_at:
            sso_configuration_record.active = True

        sso_configuration_record.configured_at = localized_utcnow()
        # Completing the orchestration process for the first time means the configuration record is now configured and
        # can be considered active. However, subsequent configurations to update the record should not be reactivated,
        # nor should the admins be renotified via email.
        if not request.query_params.get('updating_existing_record'):
            # Send a notification email to the enterprise associated with the configuration record
            send_sso_configured_email.delay(sso_configuration_record.enterprise_customer.uuid)
            sso_configuration_record.active = True

        sso_configuration_record.save()
        return Response(status=HTTP_200_OK)

    @permission_required(
        'enterprise.can_access_admin_dashboard',
        fn=lambda request, *args, **kwargs: get_customer_from_request(request)
    )
    def retrieve(self, request, *args, **kwargs):
        sso_configuration_record = fetch_configuration_record(kwargs)
        if not sso_configuration_record:
            return Response(status=HTTP_404_NOT_FOUND)
        # Make sure the requesting user is part of the enterprise customer associated with the configuration record
        # or are a staff member
        config_customer = sso_configuration_record.first().enterprise_customer
        try:
            if not check_user_part_of_customer(request.user, config_customer):
                return Response(status=HTTP_404_NOT_FOUND)
        except EnterpriseCustomerInactiveException:
            return Response(status=HTTP_403_FORBIDDEN)

        serializer = self.serializer_class(sso_configuration_record.first())
        return Response(serializer.data)

    @permission_required(
        'enterprise.can_access_admin_dashboard',
        fn=lambda request, *args, **kwargs: get_customer_from_request(request)
    )
    def list(self, request, *args, **kwargs):
        list_queryset = self.queryset

        # If the request includes a customer uuid, we require through permissioning that the requesting user is an
        # admin of that customer
        if provided_customer := request.query_params.get('enterprise_customer'):
            enterprise_customer = EnterpriseCustomer.objects.filter(pk=provided_customer)
            list_queryset = list_queryset.filter(enterprise_customer__in=enterprise_customer)

        # For non-staff users, only show the configurations that are connected to enterprise customers
        # that they are associated with
        if not request.user.is_staff:
            ent_customers = EnterpriseCustomer.objects.filter(
                enterprise_customer_users__in=EnterpriseCustomerUser.objects.filter(user_id=request.user.id)
            )
            list_queryset = list_queryset.filter(enterprise_customer__in=ent_customers)

        serializer = self.serializer_class(list_queryset, many=True)
        return Response(serializer.data, status=HTTP_200_OK)

    @permission_required(
        'enterprise.can_access_admin_dashboard',
        fn=lambda request, *args, **kwargs: get_customer_from_request(request)
    )
    def create(self, request, *args, **kwargs):
        # Force the enterprise customer to be the one associated with the user
        request_data = fetch_request_data_from_request(request)
        requesting_user_customer = request_data.get('enterprise_customer')
        if requesting_user_customer:
            try:
                enterprise_customer = EnterpriseCustomer.objects.get(uuid=requesting_user_customer)
            except EnterpriseCustomer.DoesNotExist:
                return Response(status=HTTP_403_FORBIDDEN)
            try:
                if not check_user_part_of_customer(request.user, enterprise_customer):
                    return Response(status=HTTP_403_FORBIDDEN)
            except EnterpriseCustomerInactiveException:
                return Response(status=HTTP_403_FORBIDDEN)
            request_data['enterprise_customer'] = enterprise_customer
        else:
            return Response({'error': BAD_CUSTOMER_ERROR}, status=HTTP_400_BAD_REQUEST)

        # Parse the request data to see if the metadata url or xml has changed and update the entity id if so
        sso_config_metadata_xml = None
        if request_metadata_url := request_data.get('metadata_url'):
            # If the metadata url has changed, we need to update the metadata xml
            try:
                sso_config_metadata_xml = get_metadata_xml_from_url(request_metadata_url)
            except (SsoConfigurationApiError, requests.exceptions.SSLError) as e:
                LOGGER.error(f'{BAD_IDP_METADATA_URL}{e}')
                return Response({'error': f'{BAD_IDP_METADATA_URL} {e}'}, status=HTTP_400_BAD_REQUEST)
            request_data['metadata_xml'] = sso_config_metadata_xml
        if sso_config_metadata_xml or (sso_config_metadata_xml := request_data.get('metadata_xml')):
            try:
                entity_id = fetch_entity_id_from_metadata_xml(sso_config_metadata_xml)
            except (EntityIdNotFoundError) as e:
                LOGGER.error(f'{CONFIG_UPDATE_ERROR}{e}')
                return Response({'error': f'{CONFIG_UPDATE_ERROR} {e}'}, status=HTTP_400_BAD_REQUEST)

            request_data['entity_id'] = entity_id

        try:
            with transaction.atomic():
                new_record = EnterpriseCustomerSsoConfiguration.objects.create(**request_data)
                sp_metadata_url = new_record.submit_for_configuration()
        except (TypeError, SsoOrchestratorClientError) as e:
            LOGGER.error(f'{CONFIG_CREATE_ERROR} {e}')
            return Response({'error': f'{CONFIG_CREATE_ERROR} {e}'}, status=HTTP_400_BAD_REQUEST)

        return Response({'record': new_record.pk, 'sp_metadata_url': sp_metadata_url}, status=HTTP_201_CREATED)

    @permission_required(
        'enterprise.can_access_admin_dashboard',
        fn=lambda request, *args, **kwargs: get_customer_from_request(request)
    )
    def update(self, request, *args, **kwargs):
        sso_configuration_record = fetch_configuration_record(kwargs)
        if not sso_configuration_record:
            return Response(status=HTTP_404_NOT_FOUND)
        # Make sure the requesting user is part of the enterprise customer associated with the configuration record
        # or are a staff member
        config_customer = sso_configuration_record.first().enterprise_customer
        try:
            if not check_user_part_of_customer(request.user, config_customer):
                return Response(status=HTTP_404_NOT_FOUND)
        except EnterpriseCustomerInactiveException:
            return Response(status=HTTP_403_FORBIDDEN)

        # Parse the request data to see if the metadata url or xml has changed and update the entity id if so
        request_data = fetch_request_data_from_request(request)
        sso_config_metadata_xml = None
        if request_metadata_url := request_data.get('metadata_url'):
            sso_config_metadata_url = sso_configuration_record.first().metadata_url
            if request_metadata_url != sso_config_metadata_url:
                # If the metadata url has changed, we need to update the metadata xml
                try:
                    sso_config_metadata_xml = get_metadata_xml_from_url(request_metadata_url)
                except (SsoConfigurationApiError, requests.exceptions.SSLError) as e:
                    LOGGER.error(f'{BAD_IDP_METADATA_URL}{e}')
                    return Response({'error': f'{BAD_IDP_METADATA_URL} {e}'}, status=HTTP_400_BAD_REQUEST)
                request_data['metadata_xml'] = sso_config_metadata_xml
        if request_metadata_xml := request_data.get('metadata_xml'):
            if request_metadata_xml != sso_configuration_record.first().metadata_xml:
                sso_config_metadata_xml = request_metadata_xml
        if sso_config_metadata_xml:
            try:
                entity_id = fetch_entity_id_from_metadata_xml(sso_config_metadata_xml)
                request_data['entity_id'] = entity_id
            except (EntityIdNotFoundError) as e:
                LOGGER.error(f'{CONFIG_UPDATE_ERROR}{e}')
                return Response({'error': f'{CONFIG_UPDATE_ERROR} {e}'}, status=HTTP_400_BAD_REQUEST)

        # If the request includes a customer uuid, ensure the new customer is valid
        if new_customer := request_data.get('enterprise_customer'):
            try:
                enterprise_customer = EnterpriseCustomer.objects.get(uuid=new_customer)
            except EnterpriseCustomer.DoesNotExist:
                return Response(status=HTTP_403_FORBIDDEN)
            try:
                if not check_user_part_of_customer(request.user, enterprise_customer):
                    return Response(status=HTTP_403_FORBIDDEN)
            except EnterpriseCustomerInactiveException:
                return Response(status=HTTP_403_FORBIDDEN)
        try:
            with transaction.atomic():
                needs_submitting = False
                for request_key in request_data.keys():
                    # If the requested data to update includes a field that is locked while configuring
                    if request_key in EnterpriseCustomerSsoConfiguration.fields_locked_while_configuring:
                        # If any of the provided values differ from the existing value
                        existing_value = getattr(
                            sso_configuration_record.first(), request_key, request_data[request_key]
                        )
                        if existing_value != request_data[request_key]:
                            # Indicate that the record needs to be submitted for configuration to the orchestrator
                            needs_submitting = True
                sso_configuration_record.update(**request_data)
                sp_metadata_url = ''
                if needs_submitting:
                    sp_metadata_url = sso_configuration_record.first().submit_for_configuration(
                        updating_existing_record=True
                    )
        except (TypeError, FieldDoesNotExist, ValidationError, SsoOrchestratorClientError) as e:
            LOGGER.error(f'{CONFIG_UPDATE_ERROR} {e}')
            return Response({'error': f'{CONFIG_UPDATE_ERROR} {e}'}, status=HTTP_400_BAD_REQUEST)
        serializer = self.serializer_class(sso_configuration_record.first())
        response = {'record': serializer.data}
        if sp_metadata_url:
            response['sp_metadata_url'] = sp_metadata_url
        return Response(response, status=HTTP_200_OK)

    @permission_required(
        'enterprise.can_access_admin_dashboard',
        fn=lambda request, *args, **kwargs: get_customer_from_request(request)
    )
    def destroy(self, request, *args, **kwargs):
        sso_configuration_record = fetch_configuration_record(kwargs)
        if not sso_configuration_record:
            return Response(status=HTTP_404_NOT_FOUND)
        try:
            if not check_user_part_of_customer(request.user, sso_configuration_record.first().enterprise_customer):
                return Response(status=HTTP_404_NOT_FOUND)
        except EnterpriseCustomerInactiveException:
            return Response(status=HTTP_403_FORBIDDEN)
        sso_configuration_record.update(is_removed=True)
        return Response(status=HTTP_200_OK)



# Generated by Django 3.2.12 on 2022-04-04 15:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0154_alter_systemwideenterpriseuserroleassignment_unique_together'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecustomeruser',
            name='is_relinkable',
            field=models.BooleanField(default=True, help_text='When set to False, the user cannot be relinked to the enterprise.'),
        ),
        migrations.AddField(
            model_name='historicalenterprisecustomeruser',
            name='is_relinkable',
            field=models.BooleanField(default=True, help_text='When set to False, the user cannot be relinked to the enterprise.'),
        ),
    ]



# Generated by Django 3.2.23 on 2023-11-30 18:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0194_auto_20231013_1359'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecustomerssoconfiguration',
            name='marked_authorized',
            field=models.BooleanField(blank=True, default=False, help_text='Whether admin has indicated the service provider metadata was uploaded.', null=True),
        ),
        migrations.AddField(
            model_name='historicalenterprisecustomerssoconfiguration',
            name='marked_authorized',
            field=models.BooleanField(blank=True, default=False, help_text='Whether admin has indicated the service provider metadata was uploaded.', null=True),
        ),
    ]



# Generated by Django 3.2.15 on 2022-10-31 18:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cornerstone', '0023_auto_20221021_0159'),
    ]

    operations = [
        migrations.AddField(
            model_name='cornerstoneenterprisecustomerconfiguration',
            name='deleted_at',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='historicalcornerstoneenterprisecustomerconfiguration',
            name='deleted_at',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



"""
Fields for Enterprise API serializers.
"""

import base64
import csv

from rest_framework import serializers


class Base64EmailCSVField(serializers.Field):
    """
    Serializers a Base64 encoded CSV with emails into an array of emails
    """
    def to_internal_value(self, data):
        decoded_email_csv = base64.standard_b64decode(data).decode('utf-8').splitlines()
        email_csv = list(csv.DictReader(decoded_email_csv))
        if email_csv and 'email' not in email_csv[0]:
            raise serializers.ValidationError(
                "The .csv file must have a column of email addresses, "
                "indicated by the heading 'email' in the first row."
            )
        return [email.get('email') for email in email_csv]

    def to_representation(self, value):
        return value



# Generated by Django 2.2.19 on 2021-05-17 09:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0130_lms_customer_lp_search_help_text'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecustomer',
            name='reply_to',
            field=models.EmailField(blank=True, help_text="The email address where learner's reply to enterprise emails will be delivered.", max_length=254, null=True),
        ),
        migrations.AddField(
            model_name='historicalenterprisecustomer',
            name='reply_to',
            field=models.EmailField(blank=True, help_text="The email address where learner's reply to enterprise emails will be delivered.", max_length=254, null=True),
        ),
    ]



# Generated by Django 2.2.24 on 2021-08-03 17:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('integrated_channel', '0002_learnerdatatransmissionaudit_subsection_id'),
    ]

    operations = [
        migrations.AddField(
            model_name='contentmetadataitemtransmission',
            name='content_last_changed',
            field=models.DateTimeField(blank=True, help_text='Date of the last time the enterprise catalog associated with this metadata item was updated', null=True),
        ),
    ]



# Generated by Django 2.2.15 on 2020-10-02 15:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('moodle', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='MoodleLearnerDataTransmissionAudit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('moodle_user_email', models.EmailField(help_text='The learner`s Moodle email. This must match the email on edX', max_length=255)),
                ('enterprise_course_enrollment_id', models.PositiveIntegerField(db_index=True)),
                ('course_id', models.CharField(max_length=255)),
                ('grade', models.DecimalField(blank=True, decimal_places=2, max_digits=3, null=True)),
                ('total_hours', models.FloatField(blank=True, null=True)),
                ('course_completed', models.BooleanField(default=True, help_text="The learner's course completion status transmitted to Moodle.")),
                ('completed_timestamp', models.CharField(help_text='Represents the Moodle representation of a timestamp: yyyy-mm-dd, which is always 10 characters.', max_length=10)),
                ('status', models.CharField(max_length=100)),
                ('error_message', models.TextField(blank=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
            ],
        ),
    ]



# Generated by Django 2.2.20 on 2021-06-10 08:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0133_auto_20210608_1931'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecustomerreportingconfiguration',
            name='enable_compression',
            field=models.BooleanField(default=True, help_text='Specifies whether report should be compressed. Without compression files will not be password protected or encrypted.'),
        ),
    ]



"""
Views for the ``enterprise-group`` API endpoint.
"""

from django_filters.rest_framework import DjangoFilterBackend
from edx_rbac.decorators import permission_required
from rest_framework import filters, permissions
from rest_framework.decorators import action
from rest_framework.response import Response

from django.contrib import auth
from django.db.models import Q
from django.http import Http404

from enterprise import constants, models, rules, utils
from enterprise.api.utils import get_enterprise_customer_from_enterprise_group_id
from enterprise.api.v1 import serializers
from enterprise.api.v1.views.base_views import EnterpriseReadWriteModelViewSet
from enterprise.logging import getEnterpriseLogger
from enterprise.tasks import send_group_membership_invitation_notification, send_group_membership_removal_notification
from enterprise.utils import localized_utcnow

LOGGER = getEnterpriseLogger(__name__)

User = auth.get_user_model()


class EnterpriseGroupViewSet(EnterpriseReadWriteModelViewSet):
    """
    API views for the ``enterprise-group`` API endpoint.
    """
    queryset = models.EnterpriseGroup.objects.all()
    queryset_with_removed = models.EnterpriseGroup.all_objects.all()
    permission_classes = (permissions.IsAuthenticated,)
    filter_backends = (filters.OrderingFilter, DjangoFilterBackend,)
    serializer_class = serializers.EnterpriseGroupSerializer

    def get_queryset(self, **kwargs):
        """
        - Filter down the queryset of groups available to the requesting user.
        - Account for requested filtering query params
        """
        include_deleted = self.request.query_params.get('include_deleted', False)
        if include_deleted:
            queryset = self.queryset_with_removed
        else:
            queryset = self.queryset

        if not self.request.user.is_staff:
            enterprise_user_objects = models.EnterpriseCustomerUser.objects.filter(
                user_id=self.request.user.id,
                active=True,
            )
            associated_customers = []
            for user_object in enterprise_user_objects:
                associated_customers.append(user_object.enterprise_customer)
            queryset = queryset.filter(enterprise_customer__in=associated_customers)

        if self.request.method == 'GET':
            if learner_ids := self.request.query_params.getlist('learner_ids'):
                # groups can apply to both existing and pending users
                queryset = queryset.filter(
                    Q(members__enterprise_customer_user__in=learner_ids) |
                    Q(members__pending_enterprise_customer_user__in=learner_ids),
                )
            if enterprise_uuids := self.request.query_params.getlist('enterprise_uuids'):
                queryset = queryset.filter(enterprise_customer__in=enterprise_uuids)
        return queryset

    @permission_required(
        'enterprise.can_access_admin_dashboard',
        fn=lambda request, *args, **kwargs: get_enterprise_customer_from_enterprise_group_id(kwargs['pk'])
    )
    def update(self, request, *args, **kwargs):
        """
        PATCH /enterprise/api/v1/enterprise-group/<group uuid>
        """
        if requested_customer := self.request.data.get('enterprise_customer'):
            # Essentially checking ``enterprise.can_access_admin_dashboard`` but for the customer the requester is
            # attempting to update the group record to.
            implicit_access = rules.has_implicit_access_to_dashboard(self.request.user, requested_customer)
            explicit_access = rules.has_explicit_access_to_dashboard(self.request.user, requested_customer)
            if not implicit_access and not explicit_access:
                return Response('Unauthorized', status=401)
        return super().update(request, *args, **kwargs)

    @permission_required(
        'enterprise.can_access_admin_dashboard',
        fn=lambda request, *args, **kwargs: request.POST.dict().get('enterprise_customer')
    )
    def create(self, request, *args, **kwargs):
        """
        POST /enterprise/api/v1/enterprise-group/
        """
        return super().create(request, *args, **kwargs)

    @action(detail=True, methods=['get'])
    def get_learners(self, request, *args, **kwargs):
        """
        Endpoint Location to return all learners: GET api/v1/enterprise-group/<group_uuid>/learners/

        Endpoint Location to return pending learners:
        GET api/v1/enterprise-group/<group_uuid>/learners/?pending_users_only=true

        Request Arguments:
        - ``group_uuid`` (URL location, required): The uuid of the group from which learners should be listed.

        Optional query params:
        - ``pending_users_only`` (string, optional): Specify that results should only contain pending learners
        - ``q`` (string, optional): Filter the returned members by user email and name with a provided sub-string
        - ``sort_by`` (string, optional): Specify how the returned members should be ordered. Supported sorting values
        are `memberDetails`, `memberStatus`, and `recentAction`. Ordering can be reversed by supplying a `-` at the
        beginning of the sorting value ie `-memberStatus`.
        - ``page`` (int, optional): Which page of paginated data to return.
        - ``show_removed`` (bool, optional): Include removed learners in the response.

        Returns: Paginated list of learners that are associated with the enterprise group uuid::

            {
                'count': 1,
                'next': null,
                'previous': null,
                'results': [
                    {
                        'learner_id': integer or None,
                        'pending_learner_id': integer or None,
                        'enterprise_group_membership_uuid': UUID,
                        'member_details': {
                            'user_email': string,
                            'user_name': string,
                        },
                        'recent_action': string,
                        'status': string,
                    },
                ],
            }

        """
        query_params = self.request.query_params.copy()
        is_reversed = bool(query_params.get('is_reversed', False))
        show_removed = bool(query_params.get('show_removed', False))

        param_serializers = serializers.EnterpriseGroupLearnersRequestQuerySerializer(
            data=query_params
        )

        if not param_serializers.is_valid():
            return Response(param_serializers.errors, status=400)

        user_query = param_serializers.validated_data.get('user_query')
        sort_by = param_serializers.validated_data.get('sort_by')
        pending_users_only = param_serializers.validated_data.get('pending_users_only')

        group_uuid = kwargs.get('group_uuid')
        try:
            group_object = self.get_queryset().get(uuid=group_uuid)
            members = group_object.get_all_learners(user_query,
                                                    sort_by,
                                                    desc_order=is_reversed,
                                                    pending_users_only=pending_users_only,
                                                    fetch_removed=show_removed)

            page = self.paginate_queryset(members)
            serializer = serializers.EnterpriseGroupMembershipSerializer(page, many=True)
            response = self.get_paginated_response(serializer.data)
            return response

        except models.EnterpriseGroup.DoesNotExist as exc:
            LOGGER.warning(f"group_uuid {group_uuid} does not exist")
            raise Http404 from exc

    @action(methods=['post'], detail=False, permission_classes=[permissions.IsAuthenticated])
    @permission_required(
        'enterprise.can_access_admin_dashboard',
        fn=lambda request, group_uuid: get_enterprise_customer_from_enterprise_group_id(group_uuid)
    )
    def assign_learners(self, request, group_uuid):
        """
        POST /enterprise/api/v1/enterprise-group/<group uuid>/assign_learners

        Request Arguments:
        - ``learner_emails``: List of learner emails to associate with the group. Note: only processes the first
        1000 records provided.

        Optional request data:
        - ``act_by_date`` (datetime, optional): The expiration date for the subsidy.
        - ``catalog_uuid`` (string, optional): The uuid of the catalog that is part of the subsidy.

        Returns:
        - ``records_processed``: Total number of group membership records processed.
        - ``new_learners``: Total number of group membership records associated with new pending enterprise learners
        that were processed.
        - ``existing_learners``: Total number of group membership records associated with existing enterprise learners
        that were processed.

        """
        try:
            group = self.get_queryset().get(uuid=group_uuid)
            customer = group.enterprise_customer
        except models.EnterpriseGroup.DoesNotExist as exc:
            raise Http404 from exc
        param_serializer = serializers.EnterpriseGroupRequestDataSerializer(data=request.data)
        param_serializer.is_valid(raise_exception=True)
        # act_by_date and catalog_uuid values are needed for Braze email trigger properties
        act_by_date = param_serializer.validated_data.get('act_by_date')
        catalog_uuid = param_serializer.validated_data.get('catalog_uuid')
        learner_emails = param_serializer.validated_data.get('learner_emails')
        total_records_processed = 0
        total_existing_users_processed = 0
        total_new_users_processed = 0
        for user_email_batch in utils.batch(learner_emails[: 1000], batch_size=200):
            user_emails_to_create = []
            memberships_to_create = []
            # ecus: enterprise customer users
            ecus = []
            # Gather all existing User objects associated with the email batch
            existing_users = User.objects.filter(email__in=user_email_batch)
            # Build and create a list of EnterpriseCustomerUser objects for the emails of existing Users
            # Ignore conflicts in case any of the ent customer user objects already exist
            ecu_by_email = {
                user.email: models.EnterpriseCustomerUser(
                    enterprise_customer=customer, user_id=user.id, active=True
                ) for user in existing_users
            }
            models.EnterpriseCustomerUser.objects.bulk_create(
                ecu_by_email.values(),
                ignore_conflicts=True,
            )

            # Fetch all ent customer users related to existing users provided by requester
            # whether they were created above or already existed
            ecus.extend(
                models.EnterpriseCustomerUser.objects.filter(
                    user_id__in=existing_users.values_list('id', flat=True)
                )
            )

            # Extend the list of emails that don't have User objects associated and need to be turned into
            # new PendingEnterpriseCustomerUser objects
            user_emails_to_create.extend(set(user_email_batch).difference(set(ecu_by_email.keys())))

            # Extend the list of memberships that need to be created associated with existing Users
            ent_customer_users = [
                models.EnterpriseGroupMembership(
                    activated_at=localized_utcnow(),
                    enterprise_customer_user=ecu,
                    group=group
                )
                for ecu in ecus
            ]
            total_existing_users_processed += len(ent_customer_users)
            memberships_to_create.extend(ent_customer_users)

        # Go over (in batches) all emails that don't have User objects
        for emails_to_create_batch in utils.batch(user_emails_to_create, batch_size=200):
            # Create the PendingEnterpriseCustomerUser objects
            pecu_records = [
                models.PendingEnterpriseCustomerUser(
                    enterprise_customer=customer, user_email=user_email
                ) for user_email in emails_to_create_batch
            ]
            # According to Django docs, bulk created objects can't be used in future bulk creates as the in memory
            # objects returned by bulk_create won't have PK's assigned.
            models.PendingEnterpriseCustomerUser.objects.bulk_create(pecu_records, ignore_conflicts=True)
            pecus = models.PendingEnterpriseCustomerUser.objects.filter(
                user_email__in=emails_to_create_batch,
                enterprise_customer=customer,
            )
            total_new_users_processed += len(pecus)
            # Extend the list of memberships that need to be created associated with the new pending users
            memberships_to_create.extend([
                models.EnterpriseGroupMembership(
                    pending_enterprise_customer_user=pecu,
                    group=group
                ) for pecu in pecus
            ])

        # Create all our memberships, bulk_create will batch for us.
        memberships = models.EnterpriseGroupMembership.objects.bulk_create(
            memberships_to_create, ignore_conflicts=True
        )
        total_records_processed += len(memberships)
        data = {
            'records_processed': total_records_processed,
            'new_learners': total_new_users_processed,
            'existing_learners': total_existing_users_processed,
        }
        membership_uuids = [membership.uuid for membership in memberships]
        if act_by_date and catalog_uuid:
            for membership_uuid_batch in utils.batch(membership_uuids, batch_size=200):
                send_group_membership_invitation_notification.delay(
                    customer.uuid,
                    membership_uuid_batch,
                    act_by_date,
                    catalog_uuid
                )
        return Response(data, status=201)

    @action(methods=['post'], detail=False, permission_classes=[permissions.IsAuthenticated])
    @permission_required(
        'enterprise.can_access_admin_dashboard',
        fn=lambda request, group_uuid: get_enterprise_customer_from_enterprise_group_id(group_uuid)
    )
    def remove_learners(self, request, group_uuid):
        """
        POST /enterprise/api/v1/enterprise-group/<group uuid>/remove_learners

        Required Arguments:
            - ``learner_emails``:
                List of learner emails to associate with the group.

        Optional request data:
            - ``catalog_uuid`` (string, optional): The uuid of the catalog that is part of the subsidy.

        Returns:
            - ``records_deleted``:
                Number of membership records removed
        """
        try:
            group = self.get_queryset().get(uuid=group_uuid)
            customer = group.enterprise_customer
        except models.EnterpriseGroup.DoesNotExist as exc:
            raise Http404 from exc
        param_serializer = serializers.EnterpriseGroupRequestDataSerializer(data=request.data)
        param_serializer.is_valid(raise_exception=True)

        catalog_uuid = param_serializer.validated_data.get('catalog_uuid')
        learner_emails = param_serializer.validated_data.get('learner_emails')

        records_deleted = 0
        for user_email_batch in utils.batch(learner_emails[: 1000], batch_size=200):
            existing_users = User.objects.filter(email__in=user_email_batch).values_list("id", flat=True)
            group_q = Q(group=group)
            ecu_in_q = Q(enterprise_customer_user__user_id__in=existing_users)
            pecu_in_q = Q(pending_enterprise_customer_user__user_email__in=user_email_batch)
            records_to_delete = models.EnterpriseGroupMembership.objects.filter(
                group_q & (ecu_in_q | pecu_in_q),
            )
            records_deleted += len(records_to_delete)
            records_to_delete_uuids = [record.uuid for record in records_to_delete]
            records_to_delete.delete()
            for records_to_delete_uuids_batch in utils.batch(records_to_delete_uuids, batch_size=200):
                send_group_membership_removal_notification.delay(
                    customer.uuid,
                    records_to_delete_uuids_batch,
                    catalog_uuid)
            # Woohoo! Records removed! Now to update the soft deleted records
            deleted_records = models.EnterpriseGroupMembership.all_objects.filter(
                uuid__in=records_to_delete_uuids,
            )
            deleted_records.update(
                status=constants.GROUP_MEMBERSHIP_REMOVED_STATUS,
                removed_at=localized_utcnow()
            )
        data = {
            'records_deleted': records_deleted,
        }
        return Response(data, status=200)



"""
Errors thrown by the APIs in the Enterprise application.
"""


class CodesAPIRequestError(Exception):
    """There was a problem with a request to the Codes application's APIs."""


class EnrollmentModificationException(Exception):
    """
    An exception that represents an error when modifying the state
    of an enrollment via the EnrollmentApiClient.
    """


class AdminNotificationAPIRequestError(Exception):
    """
    An exception that represents an error when creating admin notification
    read status via the NotificationReadApiClient.
    """


class LinkUserToEnterpriseError(Exception):
    """
    An error occurred while linking a user to an enterprise.
    """


class UnlinkUserFromEnterpriseError(Exception):
    """
    An error occurred while unlinking a user from an enterprise.
    """



"""
OpenAI client
"""

import openai

from django.conf import settings

openai.api_key = settings.OPENAI_API_KEY


def chat_completion(prompt, role, model='gpt-3.5-turbo'):
    """
    Use chatGPT https://api.openai.com/v1/chat/completions endpoint to generate a response.

    Arguments:
        prompt (str): ChatGPT prompt
        role (str): ChatGPT role to assume for the prompt.
        model (str): ChatGPT model to use for the prompt completion.

    Returns:
        (str): Prompt response from OpenAI.
    """
    response = openai.ChatCompletion.create(
        model=model,
        messages=[
            {'role': role, 'content': prompt},
        ]
    )

    return response['choices'][0]['message']['content']



# Generated by Django 3.2.23 on 2024-03-29 15:37

from django.db import migrations
import fernet_fields.fields


class Migration(migrations.Migration):

    dependencies = [
        ('degreed2', '0025_delete_historicaldegreed2enterprisecustomerconfiguration'),
    ]

    operations = [
        migrations.AddField(
            model_name='degreed2enterprisecustomerconfiguration',
            name='decrypted_client_id',
            field=fernet_fields.fields.EncryptedCharField(blank=True, default='', help_text='The encrypted API Client ID provided to edX by the enterprise customer to be used to make API calls to Degreed on behalf of the customer.', max_length=255, null=True, verbose_name='Encrypted API Client ID'),
        ),
        migrations.AddField(
            model_name='degreed2enterprisecustomerconfiguration',
            name='decrypted_client_secret',
            field=fernet_fields.fields.EncryptedCharField(blank=True, default='', help_text='The encrypted API Client Secret provided to edX by the enterprise customer to be used to make API calls to Degreed on behalf of the customer.', max_length=255, null=True, verbose_name='Encrypted API Client Secret'),
        ),
    ]



# Generated by Django 2.2.20 on 2021-06-08 19:31

from django.db import migrations


def remove_templates(apps, schema_editor):
    EmailTemplate = apps.get_model('enterprise', 'EnrollmentNotificationEmailTemplate')
    EmailTemplate.objects.filter(enterprise_customer=None, template_type='SELF_ENROLL').delete()
    EmailTemplate.objects.filter(enterprise_customer=None, template_type='ADMIN_ENROLL').delete()


def create_templates(apps, schema_editor):
    EmailTemplate = apps.get_model('enterprise', 'EnrollmentNotificationEmailTemplate')
    EmailTemplate.objects.get_or_create(
        plaintext_template="""
        {% load i18n %}{% if user_name %}{% blocktrans %}Dear {{ user_name }},{% endblocktrans %}{% else %}{% blocktrans %}Hi!{% endblocktrans %}{% endif %}
        {% if enrolled_in.type == "program" %}
        {% blocktrans with program_url=enrolled_in.url program_name=enrolled_in.name program_branding=enrolled_in.branding start_date=enrolled_in.start|date:"DATE_FORMAT" %}You have been enrolled in {{ program_name }}, a {{ program_branding }} program offered by {{ organization_name }}. This program begins {{ start_date }}. For more information, see the following link:

        {{ program_url }}{% endblocktrans %}{% else %}
        {% blocktrans with course_url=enrolled_in.url course_name=enrolled_in.name start_date=enrolled_in.start|date:"DATE_FORMAT" %}You have been enrolled in {{ course_name }}, a course offered by {{ organization_name }}. This course begins {{ start_date }}. For more information, see the following link:

        {{ course_url }}{% endblocktrans %}{% endif %}
        {% blocktrans with enrolled_in_name=enrolled_in.name %}
        Thanks,

        The {{enrolled_in_name}} team{% endblocktrans %}
        """,
        html_template="""
        {% load i18n %}<html>
        <body>
        <p>{% if user_name %}{% blocktrans %}Dear {{ user_name }},{% endblocktrans %}{% else %}{% blocktrans %}Hi!{% endblocktrans %}{% endif %}</p>
        <p>{% if enrolled_in.type == "program" %}
        {% blocktrans with program_url=enrolled_in.url program_name=enrolled_in.name program_branding=enrolled_in.branding start_date=enrolled_in.start|date:"DATE_FORMAT" %}You have been enrolled in <a href="{{ program_url }}">{{ program_name }}</a>, a {{ program_branding }} program offered by {{ organization_name }}. This program begins {{ start_date }}. For more information, see <a href="{{ program_url }}">{{ program_name }}</a>.{% endblocktrans %}{% else %}
        {% blocktrans with course_url=enrolled_in.url course_name=enrolled_in.name start_date=enrolled_in.start|date:"DATE_FORMAT" %}You have been enrolled in <a href="{{ course_url }}">{{ course_name }}</a>, a course offered by {{ organization_name }}. This course begins {{ start_date }}. For more information, see <a href="{{ course_url }}">{{ course_name }}</a>.{% endblocktrans %}{% endif %}
        </p>
        {% blocktrans with enrolled_in_name=enrolled_in.name %}<p>
        Thanks,
        </p>
        <p>
        The {{enrolled_in_name}} team
        </p>{% endblocktrans %}
        </body>
        </html>""",
        subject_line='',
        enterprise_customer=None,
        template_type='SELF_ENROLL',
    )

    EmailTemplate.objects.get_or_create(
        plaintext_template="""
        Great News! You've been enrolled in {{enrolled_in.name}} by {{organization_name}}

        This course is a free benefit offered especially for you, and we are excited for you to meet your learning community on edX.

        Visit this link to see and enroll in your course, {{enrolled_in.url}}

        The {{enrolled_in.name}} team
        """,
        html_template="""
        <html>
        <body>
            <table style="max-width: 800;" align="center">
                <tbody>
                    <tr>
                    <td>
                        <table width="100%" cellspacing="0" cellpadding="0" role="presentation" style="border-collapse:collapse;border-spacing:0px">
                            <tbody>
                                <tr>
                                <td align="left" valign="top" style="padding:0;Margin:0;width:560px">
                                    <img src="https://ci3.googleusercontent.com/proxy/ZAml-YgIIJEht-OBl6GqLNIUeVKLyPvs_ldFo9FShp-andj1YQxvliJXv_s_Tmh6cg1-5avJJmvXmzxbQp06sB_WUeYYkN9kzV6jtVUvYKPPjUX_8_iFJAZqsNqYakl4nQyHnl0dClFzxiaLuULSPoqGPLBfJmNOEXIBYkvKYa95640xvHwDiQ22bZ16=s0-d-e1-ft#https://appboy-images.com/appboy/communication/assets/image_assets/images/5fc568622213594dcbda2623/original.png?1606772834" width="110" height="57" border="0" style="display:block" alt="edX" id="m_-37988256656304111logo" class="CToWUd">
                                </td>
                                <td align="right" valign="top" style="padding:0;Margin:0;width:560px">
                                    <a href="https://courses.edx.org/dashboard">My Dashboard</a>
                                </td>
                                </tr>
                            </tbody>
                        </table>
                    </td>
                    </tr>
                    <tr>
                    <td align="left" bgcolor="#002b2b" style="Margin:0;padding-left:20px;padding-right:20px;padding-top:40px;padding-bottom:40px;background-color:#002b2b;background-image:url(https://ci3.googleusercontent.com/proxy/2CLnc9QL2u1L0MsUErVcQVBOz6OlDew2A5O8umOI9v7PGI3ip8YnJqYPRcexkjGsbNvNa6kUFyuHAMp7LlVBKKa7bxaKUjEO566AyX4M_6PhDtwz-QpLXLg9eQZQ93LIwP-5SbUtfxd203xXFTVBatJrN9P2hsuJSWFwd9k2pooiea6Qsg=s0-d-e1-ft#https://fzvpwi.stripocdn.email/content/guids/CABINET_4d3c6887b8ac137f656a3dd54bb5f6c8/images/53481614126118338.png);background-repeat:no-repeat;background-position:left top" background="https://ci3.googleusercontent.com/proxy/2CLnc9QL2u1L0MsUErVcQVBOz6OlDew2A5O8umOI9v7PGI3ip8YnJqYPRcexkjGsbNvNa6kUFyuHAMp7LlVBKKa7bxaKUjEO566AyX4M_6PhDtwz-QpLXLg9eQZQ93LIwP-5SbUtfxd203xXFTVBatJrN9P2hsuJSWFwd9k2pooiea6Qsg=s0-d-e1-ft#https://fzvpwi.stripocdn.email/content/guids/CABINET_4d3c6887b8ac137f656a3dd54bb5f6c8/images/53481614126118338.png">
                        <table cellpadding="0" cellspacing="0" width="100%" style="border-collapse:collapse;border-spacing:0px">
                            <tbody>
                                <tr>
                                <td align="center" valign="top" style="padding:0;Margin:0;width:560px">
                                    <table cellpadding="0" cellspacing="0" width="100%" role="presentation" style="border-collapse:collapse;border-spacing:0px">
                                        <tbody>
                                            <tr>
                                            <td align="center" style="padding:0;Margin:0">
                                                <p style="Margin:0;font-size:32px;font-family:helvetica,'helvetica neue',arial,verdana,sans-serif;line-height:48px;color:#ffffff"><strong><span class="il">Congratulations</span>, Restless Learner</strong></p>
                                            </td>
                                            </tr>
                                            <tr>
                                            <td align="center" style="padding:0;Margin:0">
                                                <p style="Margin:0;font-size:20px;font-family:arial,'helvetica neue',helvetica,sans-serif;line-height:30px;color:#ffffff">Great News! You've been <span class="il">Enrolled</span> in {{enrolled_in.name}} by {{organization_name}}</p>
                                            </td>
                                            </tr>
                                        </tbody>
                                    </table>
                                </td>
                                </tr>
                            </tbody>
                        </table>
                    </td>
                    </tr>
                    <tr>
                    <td align="left" style="padding:0;Margin:0;padding-top:20px;padding-left:20px;padding-right:20px">
                        <table cellpadding="0" cellspacing="0" width="100%" style="border-collapse:collapse;border-spacing:0px">
                            <tbody>
                                <tr>
                                <td align="center" valign="top" style="padding:0;Margin:0;width:560px">
                                    <table cellpadding="0" cellspacing="0" width="100%" role="presentation" style="border-collapse:collapse;border-spacing:0px">
                                        <tbody>
                                            <tr>
                                            <td align="center" style="padding:0;Margin:0">
                                                <p style="Margin:0;font-size:14px;font-family:arial,'helvetica neue',helvetica,sans-serif;line-height:21px;color:#333333">
                                                    This course is a free benefit offered especially for you, and we are excited for you to meet your learning community on edX.
                                                </p>
                                            </td>
                                            </tr>
                                        </tbody>
                                    </table>
                                </td>
                                </tr>
                            </tbody>
                        </table>
                    </td>
                    </tr>
                    <tr>
                    <td>
                        <table width="100%" cellspacing="0" cellpadding="0" role="presentation" style="border-collapse:collapse;border-spacing:0px">
                            <tbody>
                                <tr>
                                <td align="center" bgcolor="#ffffff" style="padding:0;Margin:0;padding-top:10px;padding-left:10px;padding-bottom:15px"><span class="m_-37988256656304111es-button-border" style="border-style:solid;border-color:#d03529;background:#d03529;border-width:0px 0px 2px 0px;display:inline-block;border-radius:0px;width:auto"><a href="{{enrolled_in.url}}" class="m_-37988256656304111es-button" style="text-decoration:none;font-family:helvetica,'helvetica neue',arial,verdana,sans-serif;font-size:18px;color:#ffffff;border-style:solid;border-color:#d03529;border-width:10px 15px;display:inline-block;background:#d03529;border-radius:0px;font-weight:normal;font-style:normal;line-height:22px;width:auto;text-align:center" target="_blank">Start my course</a></span></td>
                                </tr>
                            </tbody>
                        </table>
                    </td>
                    </tr>
                    <tr>
                    <td align="left" bgcolor="#ffffff" style="padding:0;Margin:0;padding-top:20px;padding-left:20px;padding-right:20px;background-color:#ffffff">
                        <table cellpadding="0" cellspacing="0" width="100%" style="border-collapse:collapse;border-spacing:0px">
                            <tbody>
                                <tr>
                                <td align="center" valign="top" style="padding:0;Margin:0;width:560px">
                                    <table cellpadding="0" cellspacing="0" width="100%" bgcolor="#ffffff" style="border-collapse:collapse;border-spacing:0px;background-color:#ffffff" role="presentation">
                                        <tbody>
                                            <tr>
                                            <td align="center" style="padding:0;Margin:0">
                                                <h2 style="Margin:0;line-height:29px;font-family:arial,'helvetica neue',helvetica,sans-serif;font-size:24px;font-style:normal;font-weight:normal;color:#002b2b"><strong>Share With A Friend</strong></h2>
                                            </td>
                                            </tr>
                                            <tr>
                                            <td align="center" style="padding:15px;Margin:0">
                                                <h2 style="Margin:0;line-height:17px;font-family:arial,'helvetica neue',helvetica,sans-serif;font-size:14px;font-style:normal;font-weight:normal;color:#333333">Invite a friend to take the course with you! Click below to&nbsp;share.</h2>
                                                <p style="Margin:0;font-size:14px;font-family:arial,'helvetica neue',helvetica,sans-serif;line-height:21px;color:#333333;display:none"><br></p>
                                            </td>
                                            </tr>
                                        </tbody>
                                    </table>
                                </td>
                                </tr>
                            </tbody>
                        </table>
                    </td>
                    </tr>
                    <tr>
                    <td align="center" bgcolor="#ffffff" style="padding:20px;Margin:0;background-color:#ffffff">
                        <table cellpadding="0" cellspacing="0" class="m_-37988256656304111es-left" align="left" style="border-collapse:collapse;border-spacing:0px;float:left">
                            <tbody>
                                <tr>
                                <td class="m_-37988256656304111es-m-p0r" align="center" style="padding:0;Margin:0;width:96px">
                                    <table cellpadding="0" cellspacing="0" width="100%" role="presentation" style="border-collapse:collapse;border-spacing:0px">
                                        <tbody>
                                            <tr>
                                            <td align="center" style="padding:0;Margin:0;font-size:0px"><img src="https://ci3.googleusercontent.com/proxy/2EBI4EepyVUk0cIRxzfMGmsBqU6TKjJsw7CqB8hSM0zRlLb0-BL0Y_JXhu-E6b1DiMGQ-4NHHem0qkjDgG_0xHNTYSsmJ0Rv9qBXCj2SJyT-yFlrkA7SYr4YkGg8CdkRYlkwZsQzU0DiZIDWZopRyOkPvcHWVTLwzCdJMKZcFc57HfR-eQ=s0-d-e1-ft#https://fzvpwi.stripocdn.email/content/guids/CABINET_4d3c6887b8ac137f656a3dd54bb5f6c8/images/71001614127047298.png" alt="" style="display:block;border:0;outline:none;text-decoration:none" width="35" height="35" class="CToWUd"></td>
                                            </tr>
                                            <tr>
                                            <td align="center" style="padding:0;Margin:0;padding-top:5px">
                                                <p style="Margin:0;font-size:14px;font-family:arial,'helvetica neue',helvetica,sans-serif;line-height:21px;color:#333333"><a style="font-family:arial,'helvetica neue',helvetica,sans-serif;font-size:14px;text-decoration:underline;color:#d03529" href="https://www.facebook.com/edX">Facebook</a></p>
                                            </td>
                                            </tr>
                                        </tbody>
                                    </table>
                                </td>
                                <td class="m_-37988256656304111es-hidden" style="padding:0;Margin:0;width:20px"></td>
                                </tr>
                            </tbody>
                        </table>
                        <table cellpadding="0" cellspacing="0" class="m_-37988256656304111es-left" align="left" style="border-collapse:collapse;border-spacing:0px;float:left">
                            <tbody>
                                <tr>
                                <td align="center" style="padding:0;Margin:0;width:96px">
                                    <table cellpadding="0" cellspacing="0" width="100%" role="presentation" style="border-collapse:collapse;border-spacing:0px">
                                        <tbody>
                                            <tr>
                                            <td align="center" style="padding:0;Margin:0;font-size:0px"><img src="https://ci4.googleusercontent.com/proxy/yPnmVX7UE9sXw9LbZmrtTLvkrZVy9NJurDVqU6aByY871Dj0GMzEmJZOQm5kTImdMT_qNIIvpVpTpTpboYZB6wPyZuOUKwd49j6Q7yTB9IM0DWSMdldQ4XiOHReWms7Mw7HL9rRGk1iRLwmx0IT5frUGIcmc4ulwvQcjBFSHUBcMBDrAvQ=s0-d-e1-ft#https://fzvpwi.stripocdn.email/content/guids/CABINET_4d3c6887b8ac137f656a3dd54bb5f6c8/images/19131614127052508.png" alt="" style="display:block;border:0;outline:none;text-decoration:none" width="35" height="35" class="CToWUd"></td>
                                            </tr>
                                            <tr>
                                            <td align="center" style="padding:0;Margin:0;padding-top:5px">
                                                <p style="Margin:0;font-size:14px;font-family:arial,'helvetica neue',helvetica,sans-serif;line-height:21px;color:#333333"><a style="font-family:arial,'helvetica neue',helvetica,sans-serif;font-size:14px;text-decoration:underline;color:#d03529" href="https://twitter.com/edXOnline">Twitter</a></p>
                                            </td>
                                            </tr>
                                        </tbody>
                                    </table>
                                </td>
                                <td class="m_-37988256656304111es-hidden" style="padding:0;Margin:0;width:20px"></td>
                                </tr>
                            </tbody>
                        </table>
                        <table cellpadding="0" cellspacing="0" class="m_-37988256656304111es-left" align="left" style="border-collapse:collapse;border-spacing:0px;float:left">
                            <tbody>
                                <tr>
                                <td align="center" style="padding:0;Margin:0;width:96px">
                                    <table cellpadding="0" cellspacing="0" width="100%" role="presentation" style="border-collapse:collapse;border-spacing:0px">
                                        <tbody>
                                            <tr>
                                            <td align="center" style="padding:0;Margin:0;font-size:0px"><img src="https://ci6.googleusercontent.com/proxy/GT773Vs37lIPuUvoTpVMvtQ8sSuiC9sGLGtWGNti6nwmrZGrcewvoO_zFG4XJadUV-xPRYqf9zbLEshuVcSQMZiI1yUuM5VKxTF9lLfuzDWV1ZbMPEpk2cAWEZcDAW8mD3VFxY5l0y1Gxvcc_2bB4L-ApvPHnnIw7rXB_XDe1o37B4xPaA=s0-d-e1-ft#https://fzvpwi.stripocdn.email/content/guids/CABINET_4d3c6887b8ac137f656a3dd54bb5f6c8/images/42721614127058646.png" alt="" style="display:block;border:0;outline:none;text-decoration:none" width="35" height="35" class="CToWUd"></td>
                                            </tr>
                                            <tr>
                                            <td align="center" style="padding:0;Margin:0;padding-top:5px">
                                                <p style="Margin:0;font-size:14px;font-family:arial,'helvetica neue',helvetica,sans-serif;line-height:21px;color:#333333"><a style="font-family:arial,'helvetica neue',helvetica,sans-serif;font-size:14px;text-decoration:underline;color:#d03529" href="https://www.linkedin.com/school/edx/">LinkedIn</a></p>
                                            </td>
                                            </tr>
                                        </tbody>
                                    </table>
                                </td>
                                <td class="m_-37988256656304111es-hidden" style="padding:0;Margin:0;width:20px"></td>
                                </tr>
                            </tbody>
                        </table>
                        <table cellpadding="0" cellspacing="0" class="m_-37988256656304111es-left" align="left" style="border-collapse:collapse;border-spacing:0px;float:left">
                            <tbody>
                                <tr>
                                <td align="center" style="padding:0;Margin:0;width:96px">
                                    <table cellpadding="0" cellspacing="0" width="100%" role="presentation" style="border-collapse:collapse;border-spacing:0px">
                                        <tbody>
                                            <tr>
                                            <td align="center" style="padding:0;Margin:0;font-size:0px"><img src="https://ci6.googleusercontent.com/proxy/hVT_30jYZguhmOyUAcTTZhCNDjjlivo3GAljGnyt_nNnjp0A0zcheoKw-XgM2NSD7M-CemzJEui6RIte_cH0YyJXUwCJWFpZ6j96QCB3T2UeQddtN1_VrVyQoDnAR1G82cHdJnq6Ysc-KQPvBKngjYaaPYSvcvMuHR0QS7dcnlGK35dUlA=s0-d-e1-ft#https://fzvpwi.stripocdn.email/content/guids/CABINET_4d3c6887b8ac137f656a3dd54bb5f6c8/images/23371614127063365.png" alt="" style="display:block;border:0;outline:none;text-decoration:none" width="35" height="35" class="CToWUd"></td>
                                            </tr>
                                            <tr>
                                            <td align="center" style="padding:0;Margin:0;padding-top:5px">
                                                <p style="Margin:0;font-size:14px;font-family:arial,'helvetica neue',helvetica,sans-serif;line-height:21px;color:#333333"><a style="font-family:arial,'helvetica neue',helvetica,sans-serif;font-size:14px;text-decoration:underline;color:#d03529" href="https://www.reddit.com/r/edX/">Reddit</a></p>
                                            </td>
                                            </tr>
                                        </tbody>
                                    </table>
                                </td>
                                </tr>
                            </tbody>
                        </table>
                        <table cellpadding="0" cellspacing="0" class="m_-37988256656304111es-left" align="left" style="border-collapse:collapse;border-spacing:0px;">
                            <tbody>
                                <tr>
                                <td align="left" style="padding:0;Margin:0;width:96px">
                                    <table cellpadding="0" cellspacing="0" width="100%" role="presentation" style="border-collapse:collapse;border-spacing:0px">
                                        <tbody>
                                            <tr>
                                            <td align="center" style="padding:0;Margin:0;font-size:0px"><img src="https://ci4.googleusercontent.com/proxy/9UmWL708u3GDWbf_b3dzLoavIieL1kLHnipq6vGF06ZpMlLEQe9WkHJZsNVHfBwYZhpp71tbA8EJgf2_mhGV2RwDBIiUox83T4sP6uZKF3rw6QGEtQY2Ou16eR6v39Quf2AHsnh78t6JK6PMqlxGVSVKB2WkmUX37_2jtgfIqVtOLlV1bw=s0-d-e1-ft#https://fzvpwi.stripocdn.email/content/guids/CABINET_4d3c6887b8ac137f656a3dd54bb5f6c8/images/18781614127069989.png" alt="" style="display:block;border:0;outline:none;text-decoration:none" width="35" height="35" class="CToWUd"></td>
                                            </tr>
                                            <tr>
                                            <td align="center" style="padding:0;Margin:0;padding-top:5px">
                                                <p style="Margin:0;font-size:14px;font-family:arial,'helvetica neue',helvetica,sans-serif;line-height:21px;color:#333333"><a style="font-family:arial,'helvetica neue',helvetica,sans-serif;font-size:14px;text-decoration:underline;color:#d03529" href="https://api.whatsapp.com/send?text=edX">WhatsApp</a></p>
                                            </td>
                                            </tr>
                                        </tbody>
                                    </table>
                                </td>
                                </tr>
                            </tbody>
                        </table>
                    </td>
                    </tr>
                    <tr>
                    <td>
                        <table cellpadding="0" cellspacing="0" width="100%">
                            <tbody>
                                <tr>
                                <td height="20" style="line-height:1px;font-size:1px"></td>
                                </tr>
                                <tr>
                                <td align="left" valign="top"><a href="https://business.edx.org" style="font-family:'Open Sans',Arial,sans-serif;font-size:14px;line-height:17px;text-decoration:none;color:#707070" target="_blank"><span style="color:#00262b">edX for Business</span><span style="color:#707070"> — eLearning Solutions for Your Company</span></a></td>
                                </tr>
                                <tr>
                                <td height="20" style="line-height:1px;font-size:1px"></td>
                                </tr>
                                <tr>
                                <td align="left" style="font-family:'Open Sans',Arial,sans-serif;color:#707070;font-size:14px;line-height:17px" valign="top">© 2021 edX Inc. All rights reserved.</td>
                                </tr>
                                <tr>
                                <td height="20" style="line-height:1px;font-size:1px"></td>
                                </tr>
                                <tr>
                                <td height="20" style="line-height:1px;font-size:1px"></td>
                                </tr>
                                <tr>
                                <td align="left" style="font-family:'Open Sans',Arial,sans-serif;color:#707070;font-size:14px;line-height:17px" valign="top">141 Portland St. 9th Floor, Cambridge, MA 02139</td>
                                </tr>
                                <tr>
                                <td height="30" style="line-height:1px;font-size:1px"></td>
                                </tr>
                            </tbody>
                        </table>
                    </td>
                    </tr>
                </tbody>
            </table>
        </body>
        </html>
        """,
        subject_line='',
        enterprise_customer=None,
        template_type='ADMIN_ENROLL',
    )


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0132_auto_20210608_1921'),
    ]

    operations = [
        migrations.RunPython(code=create_templates, reverse_code=remove_templates)
    ]



# Generated by Django 3.2.12 on 2022-03-23 17:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sap_success_factors', '0002_alter_sapsuccessfactorsenterprisecustomerconfiguration_display_name'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sapsuccessfactorslearnerdatatransmissionaudit',
            name='completed_timestamp',
            field=models.BigIntegerField(blank=True, null=True),
        ),
    ]



# Generated by Django 3.2.12 on 2022-05-23 16:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cornerstone', '0019_auto_20220405_2311'),
    ]

    operations = [
        migrations.AddField(
            model_name='cornerstoneenterprisecustomerconfiguration',
            name='dry_run_mode_enabled',
            field=models.BooleanField(default=False, help_text='Is this configuration in dry-run mode? (experimental)'),
        ),
        migrations.AddField(
            model_name='historicalcornerstoneenterprisecustomerconfiguration',
            name='dry_run_mode_enabled',
            field=models.BooleanField(default=False, help_text='Is this configuration in dry-run mode? (experimental)'),
        ),
    ]
